// Main class
// Assignment 2
// CSC 375 - Lea
// Jacob Peck - 20111024

// FindBugs reports 0 bugs as of 20111024

import java.io.File;
import java.io.FileNotFoundException;
import java.util.Scanner;
import java.util.Map;
import java.util.HashMap; // wrapped in a RWL is method 1
import java.util.concurrent.ConcurrentHashMap; // method 2
import jsr166e.ConcurrentHashMapV8; // method 3
import java.util.concurrent.CountDownLatch;

public class Main {
  
  static Map<String,String> dataStore;
  static Thread[] threads;
  static Worker[] workers;
  static ReadWriteLock lock = new ReadWriteLock();  
  
  static CountDownLatch startLatch;
  static CountDownLatch stopLatch;
  
  public static void main(String[] args) {
    int threadCount = 0;
    int method = 0; // 0 = RWL HashMap, 1 = ConcurrentHashMap, 2 = ConcurrentHashMapV8
    
    //// getopts
    if (args.length != 2) {
      System.out.println("Usage: java Main <number of threads> <method>\n" +
                         "where <number of threads> is the number of threads to use\n" +
                         "and <method> is either 0 (RWL), 1(ConcurrentHashMap), or 2 (ConcurrentHashMapV8)");
      System.exit(0);
    } else {
      threadCount = Integer.parseInt(args[0]);
      if (threadCount < 1) {
        System.out.println("<number of threads> must be at least 1.");
        System.exit(0);
      }
      method = Integer.parseInt(args[1]);
      if (method < 0 || method > 2) {
        System.out.println("<method> must be either 0 (RWL), 1 (ConcurrentHashMap)," +
                           " or 2 (ConcurrentHashMapV8)");
        System.exit(0);
      }
    }
    
    //// initialize the datastore
    System.out.println("Initializing the datastore... using method " + method); 
    if (method == 0) dataStore = new HashMap<String, String>();
    else if (method == 1) dataStore = new ConcurrentHashMap<String, String>();
    else dataStore = new ConcurrentHashMapV8<String, String>();
    
    System.out.println("Populating the datastore with initial data from data.dat...");
    try {
      Scanner sc = new Scanner(new File("data.dat"));
      while(sc.hasNextLine()) {
        String[] line = sc.nextLine().split(":");
        if (line[0].startsWith("#")) continue; // comment line
        dataStore.put(line[0],line[1]); // even if RWL method, no threading going on, so it's safe to operate unlocked
      }
      System.out.println("Datastore successfully initialized.");
    } catch (FileNotFoundException ex) {
      System.out.println("No data file found.  Datastore initialized to empty.");
    }
    
    //// make the workers
    System.out.println("Initializing workers...");
    
    // signals
    startLatch = new CountDownLatch(1);
    stopLatch = new CountDownLatch(threadCount);
    
    // threads
    threads = new Thread[threadCount];
    workers = new Worker[threadCount];
    
    boolean ignoreRWL = (method != 0);
    for (int i = 0; i < threadCount; i++)
      workers[i] = new Worker(ignoreRWL,dataStore,lock);
      
    //// create threads and start, awaiting the start signal
    System.out.println("Initializing and starting threads...");
    for (int i = 0; i < threadCount; i++)
      threads[i] = makeThread(workers[i]);
    for (int i = 0; i < threadCount; i++)
      threads[i].start();
      
    //// start the system
    System.out.println("System running...");   
    long endTime;
    long startTime = System.nanoTime();
    startLatch.countDown();
    
    try {
      stopLatch.await();
    } catch (InterruptedException ex) {
    } finally {
      endTime = System.nanoTime();
    }
    
    long totalTime = endTime - startTime;
    long numOps = 1000 * threadCount;
    long avgOperationTime = totalTime/numOps;
    System.out.println("System finished.\n  Total time elapsed (ns): " + totalTime);
    System.out.println("  Number of operations performed: " + numOps);
    System.out.println("  Average operation time (ns): " + avgOperationTime);
    
    System.exit(0);
 
  }
  
  // Returns a new Thread, consisting of a Worker continually doing something
  // 10% write,90% read
  private static Thread makeThread(final Worker worker) {
    Runnable runloop = new Runnable() {
      public void run() {
        try{
          startLatch.await();
          for(int i = 0; i < 1000; i++)
            worker.doSomething();
          stopLatch.countDown();
        } catch (InterruptedException ex) {return;}
      }
    };
    return new Thread(runloop);
  }
}
